import os
from urllib.request import url2pathname

from rsrc.models import (Resource,
                         URL)

from .models import (Directory,
                     File)


def deserialize_path(string: str) -> Resource:
    if os.path.isdir(string):
        return Directory.from_string(string)
    return File.from_string(string)


def deserialize_url(string: str) -> Resource:
    url = URL.from_string(string)
    return deserialize_path(url2pathname(url.authority + url.path_string))



import os
import shutil
from errno import (EINVAL,
                   EISDIR,
                   ENOTDIR)
from pathlib import Path
from stat import (S_ISDIR,
                  S_ISREG)
from typing import (Any,
                    IO,
                    Iterator,
                    Type)

from memoir import cached
from reprit import seekers
from reprit.base import generate_repr
from rsrc.models import (Base,
                         Container,
                         Stream,
                         URL)

from .hints import Domain


def path_to_resource(path: Path) -> Base:
    if path.is_dir():
        return Directory(path)
    return File(path)


class Directory(Container):
    __slots__ = ('_path',)

    def __init__(self, path: Path) -> None:
        self._path = path

    def __iter__(self) -> Iterator[Base]:
        yield from map(path_to_resource, self._path.iterdir())

    def __hash__(self) -> int:
        return hash(self._path)

    def __eq__(self, other: Base) -> bool:
        if not isinstance(other, Base):
            return NotImplemented
        if not isinstance(other, Directory):
            return False
        return self._path == other._path

    @cached.property_
    def url(self) -> URL:
        return URL.from_string(self._path.as_uri())

    def exists(self) -> bool:
        try:
            stat = self._path.stat()
        except FileNotFoundError:
            return False
        else:
            if S_ISDIR(stat.st_mode):
                return True
            raise _to_os_error(error_code=ENOTDIR,
                               location_string=str(self))

    def join(self, part: str, *parts: str) -> Base:
        return path_to_resource(self._path.joinpath(part, *parts))

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)

    def __str__(self) -> str:
        return str(self._path)

    @classmethod
    def from_string(cls: Type[Domain], string: str) -> Domain:
        path = Path(string)
        if path.is_file():
            raise ValueError('Invalid argument: "{argument}" '
                             'supposed to be a path to directory, '
                             'but found file.'
                             .format(argument=string))
        return cls(path)


class File(Stream[str]):
    __slots__ = ('_path',)

    def __init__(self, path: Path) -> None:
        self._path = path

    def __hash__(self) -> int:
        return hash(self._path)

    def __eq__(self, other: Base) -> bool:
        if not isinstance(other, Base):
            return NotImplemented
        if not isinstance(other, File):
            return False
        return self._path == other._path

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)

    def __str__(self) -> str:
        return str(self._path)

    def exists(self) -> bool:
        try:
            stat = self._path.stat()
        except FileNotFoundError:
            return False
        else:
            mode = stat.st_mode
            if S_ISREG(mode):
                return True
            raise _to_os_error(error_code=EISDIR if S_ISDIR(mode) else EINVAL,
                               location_string=str(self))

    def open(self,
             *,
             binary_mode: bool = False,
             encoding: str = None) -> IO:
        return self._path.open(mode='rb' if binary_mode else 'r',
                               encoding=encoding)

    def send(self, destination: Base, **kwargs: Any) -> None:
        if not isinstance(destination, Stream):
            raise TypeError('Unsupported destination type: {type}.'
                            .format(type=type(destination)))
        if isinstance(destination, File):
            shutil.copy(str(self), str(destination))
        else:
            destination.receive(self, **kwargs)

    def receive(self, source: Base, **kwargs: Any) -> None:
        if not isinstance(source, Stream):
            raise TypeError('Unsupported destination type: {type}.'
                            .format(type=type(source)))
        if isinstance(source, File):
            shutil.copy(str(source), str(self))
        else:
            source.send(self, **kwargs)

    @cached.property_
    def url(self) -> URL:
        return URL.from_string(self._path.as_uri())

    @classmethod
    def from_string(cls: Type[Domain], string: str) -> Domain:
        path = Path(string)
        if path.is_dir():
            raise ValueError('Invalid argument: "{argument}" '
                             'supposed to be a path to file, '
                             'but found directory.'
                             .format(argument=string))
        return cls(path)


def _to_os_error(*,
                 error_code: int,
                 location_string: str) -> OSError:
    return OSError(error_code, os.strerror(error_code), location_string)



import sys
from pathlib import Path
from string import ascii_lowercase

from rsrc import plugins
from setuptools import (find_packages,
                        setup)

import rsrc_local
from rsrc_local import base

plugins_entry_points = [
    plugins.to_entry_point(id_=plugins.to_id(''),
                           module_name=base.__name__,
                           function_name=base.deserialize_path.__qualname__),
    plugins.to_entry_point(id_=plugins.to_id('file'),
                           module_name=base.__name__,
                           function_name=base.deserialize_url.__qualname__),
]
if sys.platform == 'win32':
    # workaround for supporting absolute Windows paths with drive letters
    # which may be considered as a scheme during URL parsing
    for drive_letter in ascii_lowercase:
        plugins_entry_points.append(plugins.to_entry_point(
                id_=plugins.to_id(drive_letter),
                module_name=base.__name__,
                function_name=base.deserialize_path.__qualname__))

project_base_url = 'https://github.com/lycantropos/rsrc_local/'

install_requires = [
    'memoir>=0.0.1',
    'reprit>=0.0.1',
    'rsrc>=0.0.2',
]
setup_requires = [
    'pytest-runner>=4.2',
]
tests_require = Path('requirements-tests.txt').read_text()

setup(name=rsrc_local.__name__,
      packages=find_packages(exclude=('tests', 'tests.*')),
      version=rsrc_local.__version__,
      description=rsrc_local.__doc__,
      long_description=Path('README.md').read_text(encoding='utf-8'),
      long_description_content_type='text/markdown',
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url=project_base_url,
      download_url=project_base_url + 'archive/master.zip',
      python_requires='>=3.5.3',
      entry_points={plugins.__name__: plugins_entry_points},
      install_requires=install_requires,
      setup_requires=setup_requires,
      tests_require=tests_require)
